/***
 * Juliac
 * Copyright (C) 2008-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.compile.jdt;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.util.Util;
import org.objectweb.fractal.juliac.SourceFile;

/**
 * A basic implementation of {@link ICompilationUnit}. This class originates
 * from <a href="http://spoon.gforge.inria.fr">Spoon</a>.
 * 
 * @author Renaud Pawlak <renaud.pawlak@isep.fr>
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.3
 * @see spoon.support.util.BasicCompilationUnit
 */
public class BasicCompilationUnit implements ICompilationUnit {
    
    protected char[] contents;
    protected char[] fileName;
    protected char[][] packageName;
    protected char[] mainTypeName;
    protected String encoding;

    public BasicCompilationUnit(
        char[] contents, char[][] packageName, String fileName ) {
        
        this.contents = contents;
        this.fileName = fileName.toCharArray();
        this.packageName = packageName;
    }

    public char[] getContents() {
        if (this.contents != null)
            return this.contents; // answer the cached source

        // otherwise retrieve it
        try {
            return Util.getFileCharContent(new File(new String(this.fileName)),
                    this.encoding);
        } catch (IOException e) {
            // could not read file: returns an empty array
        }
        return CharOperation.NO_CHAR;
    }

    /**
     * @see org.eclipse.jdt.internal.compiler.env.IDependent#getFileName()
     */
    public char[] getFileName() {
        return this.fileName;
    }

    public char[] getMainTypeName() {
        if (this.mainTypeName == null) {
            int start = CharOperation.lastIndexOf('/', this.fileName) + 1;
            if (start == 0
                    || start < CharOperation.lastIndexOf('\\', this.fileName))
                start = CharOperation.lastIndexOf('\\', this.fileName) + 1;
            int separator = CharOperation.indexOf('|', this.fileName) + 1;
            if (separator > start) // case of a .class file in a default
                // package in a jar
                start = separator;

            int end = CharOperation.lastIndexOf('$', this.fileName);
            if (end == -1 || !Util.isClassFileName(this.fileName)) {
                end = CharOperation.lastIndexOf('.', this.fileName);
                if (end == -1)
                    end = this.fileName.length;
            }

            this.mainTypeName = CharOperation.subarray(this.fileName, start,
                    end);
        }
        return this.mainTypeName;
    }

    public char[][] getPackageName() {
        return this.packageName;
    }

    @Override
    public String toString() {
        return "CompilationUnit: " + new String(this.fileName); //$NON-NLS-1$
    }

    /**
     * Construct a {@link ICompilationUnit}-indexed map of {@link SourceFile}s.
     * 
     * @param sfs  the source files
     * @return     the map
     * @since 2.2.1
     */
    static Map<ICompilationUnit,SourceFile>
        getICompilationUnitMap( List<SourceFile> sfs ) throws IOException {
        
        Map<ICompilationUnit,SourceFile> units =
            new HashMap<ICompilationUnit,SourceFile>();
        for (SourceFile sf : sfs) {
            ICompilationUnit unit = getUnit(sf);
            units.put(unit,sf);
        }
        return units;
    }

    /**
     * Return the compilation unit corresponding to the specified source file.
     */
    private static ICompilationUnit getUnit( SourceFile sf )
    throws IOException {
    
        String qname = sf.getQname();
        String fileName = sf.getName();
        char[] content = sf.getContent();
        
        String[] tmp = qname.split("[.]");
        char[][] packageName = new char[tmp.length - 1][];
    
        for (int i = 0; i < tmp.length - 1; i++) {
            packageName[i] = tmp[i].toCharArray();
        }
    
        ICompilationUnit unit =
            new BasicCompilationUnit( content, packageName, fileName );
        return unit;
    }
}
