/***
 * Juliac
 * Copyright (C) 2008-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributors: Loris Bouzonnet, Philippe Merle
 */

package org.objectweb.fractal.juliac.compile.jdt;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.IErrorHandlingPolicy;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileConstants;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.SourceFile;
import org.objectweb.fractal.juliac.api.CompileSupportItf;
import org.objectweb.fractal.juliac.conf.JDKLevel;
import org.objectweb.fractal.juliac.conf.JuliacConfig;

/**
 * This class implements a compilation service based on the Eclipse JDT
 * compiler. This service is required by Juliac.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Loris Bouzonnet <Loris.Bouzonnet@bull.net>
 * @since 1.3
 */
public class CompileSupportImpl implements CompileSupportItf {

    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

	private Juliac jc;
	
	final public static Class<CompileSupportItf> SERVICE_TYPE =
		CompileSupportItf.class;
	
    public void init( Juliac jc ) {
    	this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(SERVICE_TYPE,this);
}
    
    public void close( Juliac jc ) {
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(SERVICE_TYPE,this);
    }
    
    
    // -----------------------------------------------------------------------
    // Implementation of the CompileSupportItf interface
    // -----------------------------------------------------------------------

    /**
     * Compile the specified source files and dump the resulting bytecode files
     * into the specified directory.
     * 
     * @param sfs       the source files
     * @param classDir  the directory for storing bytecode files
     * @return 
     * 		the paths relative to <code>classDir</code> of compiled bytecode
     * 		files 
     * @throws IOException  if an error occurs during the compilation
     */
    public List<String> compile( List<SourceFile> sfs, File classDir )
    throws IOException {
        
        /*
         * Compile.
         */
        JDTCompiler jdtcompiler = getJDTCompiler(sfs);
        compile(jdtcompiler,sfs);
        
        /*
         * Dump compiled files to disk.
         */
        String path = classDir.getAbsolutePath();
        List<String> outputFileNames = new ArrayList<String>();
        List<ClassFile> files = jdtcompiler.getClassFiles();
        for (ClassFile f : files) {
            String name =
                new String(f.fileName()).replace('/',File.separatorChar) +
                ".class";
            ClassFile.writeToDisk(true,path,name,f);
            String filename = new String(f.fileName())+".class";
            outputFileNames.add(filename);
        }
        
        return outputFileNames;
    }

    
    // -----------------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------------

    /**
     * Compile the specified source files. This method is invoked by the Spoon
     * module.
     * 
     * @param sfs  the source files to compile
     * @param reportErrorsOnly
     * 		<code>true</code> if only errors must be reported
     * @return  the corresponding compilation unit declarations
     * @throws IOException  if an error occurs during the compilation
     * @since 2.2.1
     */
    public List<CompilationUnitDeclaration> compile( List<SourceFile> sfs )
    throws IOException {
        JDTCompiler jdtcompiler = getJDTCompiler(sfs);
        return compile(jdtcompiler,sfs);
    }

    /**
     * @since 2.2.1
     */
    private List<CompilationUnitDeclaration> compile(
        JDTCompiler jdtcompiler, List<SourceFile> sfs )
    throws IOException {

        /*
         * Compile.
         * 
         * Provide the map of compilation units to the requestor in case an
         * error occurs during the compilation to be able to trace back the
         * error to the right source file.
         */
        Map<ICompilationUnit,SourceFile> units =
            BasicCompilationUnit.getICompilationUnitMap(sfs);
        jdtcompiler.compile(units);
        
        /*
         * Throw IOException in case of compilation error.
         */
        boolean hasError = jdtcompiler.hasError();
        if( hasError ) {
        	// Put the compilation error message into the exception
        	// in order to inform the caller on the compilation error message.
            String msg = "Compilation error:\n" + jdtcompiler.getErrorMessage();
            throw new IOException(msg);
        }
        
        List<CompilationUnitDeclaration> unitds = jdtcompiler.getUnits();
        return unitds;
    }

    /**
     * @since 2.2.1
     */
    private JDTCompiler getJDTCompiler( List<SourceFile> sfs ) {
        
        JuliacConfig jconf = jc.getJuliacConfig();
        JDKLevel sourceLevel = jconf.getSourceLevel();
        JDKLevel targetLevel = jconf.getTargetLevel();
        ClassLoader classLoader = jconf.getClassLoader();

        CompilerOptions options = new CompilerOptions();
        options.sourceLevel = levels.get(sourceLevel);
        options.targetJDK = levels.get(targetLevel);
        /*
         * TODO: Investigate whether we need an option to set the compliance
         * level to a value different from that of the source level.
         */
        options.complianceLevel = options.sourceLevel;
        
        /*
         * Contribution by Loris.
         */
        // Generate debug info for debugging the output.
        options.produceDebugAttributes = ClassFileConstants.ATTR_VARS
            | ClassFileConstants.ATTR_LINES | ClassFileConstants.ATTR_SOURCE;
        // Tricks like "boolean stopHere = true;" depend on this setting.
        options.preserveAllLocalVariables = true;
        
        List<String> sources = new ArrayList<String>(sfs.size());

        for (SourceFile sf : sfs) {
			String qname = sf.getQname();
			sources.add(qname);
        }
        INameEnvironment environment =
        	new NameEnvironmentImpl(classLoader,sources);
        IErrorHandlingPolicy policy = getErrorHandlingPolicy();

        JDTCompiler jdtcompiler =
        	new JDTCompiler(jc,environment,policy,options);
        return jdtcompiler;
    }
    
    /**
     * @since 2.2.2
     */
    final private static Map<JDKLevel,Long> levels =
        new HashMap<JDKLevel,Long>() {
            private static final long serialVersionUID = -6067687147064920146L;
        {
            put(JDKLevel.JDK1_1,ClassFileConstants.JDK1_1);
            put(JDKLevel.JDK1_2,ClassFileConstants.JDK1_2);
            put(JDKLevel.JDK1_3,ClassFileConstants.JDK1_3);
            put(JDKLevel.JDK1_4,ClassFileConstants.JDK1_4);
            put(JDKLevel.JDK1_5,ClassFileConstants.JDK1_5);
            put(JDKLevel.JDK1_6,ClassFileConstants.JDK1_6);
            put(JDKLevel.JDK1_7,ClassFileConstants.JDK1_7);
        }};
    
    /**
     * @since 2.2.1
     */
    private IErrorHandlingPolicy getErrorHandlingPolicy() {
        return new IErrorHandlingPolicy() {
            public boolean proceedOnErrors() {
                return true; // stop if there are some errors
            }
            public boolean stopOnFirstError() {
                return false;
            }
        };
    }
}
